import click

from enlightenme.commands.source_command import SourceCommand
from enlightenme.news.news_formatters import NewsFormatter

ENLIGHTENME_HELP = "collect news/updates from your favorite tech channels"


@click.group(name="enlightenme",
             help=ENLIGHTENME_HELP,
             invoke_without_command=True)
@click.pass_context
def cli(ctx: click.Context):
    if not ctx.invoked_subcommand:
        click.echo("""
Welcome to Enlighten Me!
------------------------
%s

See enlightenme --help for more details.""" % ENLIGHTENME_HELP)


@cli.group(cls=SourceCommand, help=SourceCommand.help)
@click.option('--format', '-f', default=NewsFormatter.DEFAULT_FORMAT,
              type=click.Choice(NewsFormatter.FORMAT_OPTIONS),
              help="Displays news in a format")
@click.option('--output', '-o', required=False,
              help="Write to FILE instead of stdout")
@click.option('--keywords', '-k', required=False,
              help="Topics or Keywords that interest you (e.g. python,golang)")
@click.pass_context
def source(ctx: click.Context, format: str = None,
           output: str = None, keywords: str = None):
    pass



from datetime import datetime
from typing import List

import click
import praw

from enlightenme.news.news import News
from enlightenme.sources import Source


class RedditSource(Source):
    NUMBER_OF_REDDITS = 10
    HELP = "Source: Reddit (http://reddit.com/)"

    def __init__(self, client_id: str = None,
                 client_secret: str = None) -> None:
        self._client_id = client_id
        self._client_secret = client_secret
        self.reddit = praw.Reddit(client_id=client_id,
                                  user_agent='enlightenme',
                                  client_secret=self._client_secret)

    @classmethod
    def name(cls):
        return 'reddit'

    @classmethod
    def params(cls) -> List[click.Parameter]:
        client_help = "See: https://github.com/reddit-archive/" + \
                      "reddit/wiki/OAuth2-Quick-Start-Example#first-steps"
        return [
            click.Option(('--client-id', '-c'),
                         envvar='REDDIT_CLIENT_ID',
                         required=True,
                         prompt=True,
                         hide_input=True,
                         help=client_help),
            click.Option(('--client-secret', '-s'),
                         envvar='REDDIT_CLIENT_SECRET',
                         required=True,
                         prompt=True,
                         hide_input=True,
                         help=client_help)
        ]

    def fetch(self, keywords: List = None) -> List[News]:
        subreddits = '+'.join(keywords) if keywords else 'all'
        hot_reddits = self.reddit.subreddit(subreddits). \
            hot(limit=RedditSource.NUMBER_OF_REDDITS)
        return [
            News(title=submission.title,
                 published_at=datetime.fromtimestamp(submission.created_utc),
                 url=submission.url,
                 body=submission.selftext,
                 tags=[str(submission.subreddit)]
                 )
            for submission in hot_reddits]



import importlib
import pkgutil
import sys


def camel_case(word):
    return "".join([group[0].upper() + group[1:] for group in word.split("-")])


def import_submodules(package_name):
    """ Import all submodules of a module, recursively

    :param package_name: Package name
    :type package_name: str
    :rtype: dict[types.ModuleType]
    """

    package = sys.modules[package_name]
    return {
        name: importlib.import_module(package_name + '.' + name)
        for loader, name, is_pkg in pkgutil.walk_packages(package.__path__)
    }



from datetime import datetime


class News:
    def __init__(self, title, published_at: datetime = None,
                 body: str = None, url: str = None, tags=None):
        if tags is None:
            tags = []
        self.title = title
        self.published_at = published_at if published_at else datetime.now()
        self.body = body
        self.url = url
        self.tags = tags

    def contains_any_keywords(self, keywords):
        for keyword in keywords:
            if self.has_any_keyword(keyword):
                return True
        return False

    def has_any_keyword(self, keyword: str) -> bool:
        upper_case_keyword = keyword.upper()
        return (self.title and upper_case_keyword in self.title.upper()) or \
               (self.tags and upper_case_keyword in
                [tag.upper() for tag in self.tags]) or \
               (self.body and upper_case_keyword in self.body.upper())



from abc import ABCMeta, abstractclassmethod, abstractmethod
from typing import List, Optional

import click

from enlightenme.news.news import News
from enlightenme.utils import import_submodules


class Source:
    __metaclass__ = ABCMeta
    HELP = None

    @abstractclassmethod
    def name(self):
        pass

    @classmethod
    def get_all_sources(cls) -> List[str]:
        return [kls.name() for kls in cls._sub_classes()
                if kls.name() is not None]

    @classmethod
    def get_source(cls, source_name: str) -> Optional[type]:
        try:
            return next(kls for kls in cls._sub_classes()
                        if kls.name() == source_name)
        except StopIteration:
            return None

    @classmethod
    def params(cls) -> List[click.Parameter]:
        return []

    @abstractmethod
    def fetch(self, keywords: List = None) -> List[News]:
        raise NotImplementedError()

    @classmethod
    def _sub_classes(cls) -> List:
        import_submodules('enlightenme.sources')
        return cls.__subclasses__()



__title__ = 'enlightenme'
__description__ = 'LI for curating and reading news/updates.'
__url__ = 'https://github.com/singhgarima/enlightenme'
__version__ = '0.0.5'
__author__ = 'Garima Singh'
__author_email__ = 'igarimasingh+enlightenme@gmail.com'
__license__ = 'MIT License'
__copyright__ = 'Copyright 2018 Garima Singh'



import click

from enlightenme.news import news_manager, news_output
from enlightenme.sources import Source


class SourceCommand(click.Group):
    help = "helps getting latest news from specified sources"

    def list_commands(self, ctx):
        return [name for name in Source.get_all_sources()]

    def get_command(self, ctx: click.Context, cmd_name):
        if cmd_name not in self.list_commands(ctx):
            ctx.fail("Invalid source %s supplied. See --help" % cmd_name)

        @click.pass_context
        def callback(cmd_ctx: click.Context, *args, **kwargs):
            keywords = ctx.params.get('keywords', "")
            format_type = ctx.params.get('format')
            output = ctx.params.get('output')

            keywords = keywords.split(",") if keywords else None

            news_fetchers = news_manager.NewsManager(
                cmd_name,
                source_params=cmd_ctx.params,
                format_type=format_type,
                keywords=keywords)

            click.echo("Fetching news from source: %s" % cmd_name)
            content = news_fetchers.fetch_and_format()
            news_output.NewsOutput(content).write_to(file_name=output)

        source_class = Source.get_source(cmd_name)
        return click.Command(name=cmd_name,
                             callback=callback,
                             params=source_class.params(),
                             help=source_class.HELP,
                             short_help=source_class.HELP)



from typing import List, Dict

from enlightenme.news import news_formatters
from enlightenme.news.news_formatters import NewsFormatter
from enlightenme.sources import Source


class NewsManager:
    def __init__(self, source_name: str,
                 source_params: Dict = None,
                 keywords: List = None,
                 format_type=NewsFormatter.DEFAULT_FORMAT):
        self._source_name = source_name
        self._source_params = source_params if source_params else {}
        self._format_type = format_type
        self._keywords = keywords
        self._news_list = []

    def fetch_and_format(self) -> str:
        self._fetch()
        return self._format()

    def _fetch(self):
        source_class = Source.get_source(self._source_name)
        __import__(source_class.__module__, fromlist=[source_class.__name__])
        source_object = source_class(**self._source_params)
        self._news_list = source_object.fetch(keywords=self._keywords)

    def _format(self) -> str:
        kls = news_formatters.NewsFormatter.formatter_for_format(
            self._format_type)
        formatter = kls(self._news_list)
        return formatter.format()



import click


class NewsOutput:
    def __init__(self, content):
        self._content = content

    def write_to(self, file_name: str = None) -> None:
        if file_name is None:
            self.write_to_stdout()
        else:
            self.write_to_file(file_name)

    def write_to_stdout(self):
        click.echo(self._content)

    def write_to_file(self, file_name):
        with open(file_name, 'w') as f:
            f.write(self._content)



import json
import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

about = {}
about_filename = os.path.join(here, 'enlightenme', '__version__.py')
with open(about_filename, 'r', encoding='utf-8') as f:
    exec(f.read(), about)

long_description = None
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

install_requires = []
tests_require = []

with open('Pipfile.lock') as fd:
    lock_data = json.load(fd)
    install_requires = [
        package_name + package_data['version']
        for package_name, package_data in lock_data['default'].items()
    ]
    tests_require = [
        package_name + package_data['version']
        for package_name, package_data in lock_data['develop'].items()
    ]

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    license=about['__license__'],

    install_requires=install_requires,
    tests_require=tests_require,

    packages = find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    entry_points='''
        [console_scripts]
        enlightenme=enlightenme.enlightenme:cli
    ''',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
)



from datetime import datetime
from typing import List, Optional

import requests

from enlightenme.news.news import News
from enlightenme.sources import Source


class HackerNewsSource(Source):
    URL = "https://hacker-news.firebaseio.com/"

    NUMBER_OF_STORIES = 10
    MAX_STORIES_TO_LOOP = 100

    HELP = "Source: hacker-news (https://news.ycombinator.com/)"

    @classmethod
    def name(cls):
        return "hacker-news"

    def __init__(self) -> None:
        super(HackerNewsSource, self).__init__()
        self._story_ids = []

    def fetch(self, keywords: List = None) -> List[News]:
        self._story_ids = self._fetch_new_stories()[0:self.MAX_STORIES_TO_LOOP]
        if keywords:
            return self._fetch_latest_stories_with_keywords(keywords)
        else:
            return self._fetch_latest_stories(number=self.NUMBER_OF_STORIES)

    def _fetch_new_stories(self):
        response = requests.get(self.URL + "v0/newstories.json")
        if response.status_code == 200:
            return response.json()
        response.raise_for_status()

    def _fetch_latest_stories(self, number=20) -> List[News]:
        latest_news = []
        for story_id in self._story_ids[0:number]:
            news = HackerNewsStory(story_id).fetch()
            if news is not None:
                latest_news.append(news)
        return latest_news

    def _fetch_latest_stories_with_keywords(self, keywords: List) \
            -> List[News]:
        interesting_news = []
        for index, story_id in enumerate(self._story_ids):
            news = HackerNewsStory(story_id).fetch()
            if news:
                if news.contains_any_keywords(keywords):
                    interesting_news.append(news)
                if len(interesting_news) == self.NUMBER_OF_STORIES:
                    break
        return interesting_news


class HackerNewsStory:
    def __init__(self, story_id: int, story_details: dict = None):
        self._story_id = story_id
        self._story_details = story_details

    def fetch(self) -> Optional[News]:
        response = requests.get(HackerNewsSource.URL +
                                "v0/item/%s.json" % str(self._story_id))
        if response.status_code == 200:
            self._story_details = response.json()
            if self._story_details is not None:
                return self._create_new_from_story_details()
            return None
        response.raise_for_status()

    def _create_new_from_story_details(self) -> News:
        published_at = datetime.fromtimestamp(self._story_details.get("time"))
        return News(self._story_details.get("title"),
                    published_at,
                    self._story_details.get("text"),
                    self._story_details.get("url"))



import csv
import io
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from typing import List

from enlightenme.news.news import News
from enlightenme.utils import camel_case

DATE_TIME_STR_FORMAT = '%Y-%m-%dT%H:%M:%SZ'


class NewsFormatter:
    __metaclass__ = ABCMeta

    DEFAULT_FORMAT = "list"
    FORMAT_OPTIONS = ['list', 'csv']

    def __init__(self, news_list: List[News]):
        self._news_list = news_list

    @property
    def news_list(self):
        return self._news_list

    @news_list.setter
    def news_list(self, value: List[News]):
        self._news_list = value

    @abstractmethod
    def format(self) -> str:
        raise NotImplementedError

    @classmethod
    def formatter_for_format(cls, format_type: str):
        return eval("%sNewsFormatter" % camel_case(format_type))


class ListNewsFormatter(NewsFormatter):
    def format(self) -> str:
        output = "\n"
        for index, news in enumerate(self.news_list):
            output += "\t%d. Title: %s\n" % (index + 1, news.title)
            published_at = news.published_at.strftime(DATE_TIME_STR_FORMAT)
            output += "\t    Published At: %s\n" % published_at

            if news.url:
                output += "\t    URL: %s\n" % news.url
            if news.tags:
                output += "\t    Tags: %s\n" % ", ".join(news.tags)
        return output


class CsvNewsFormatter(NewsFormatter):
    def format(self) -> str:
        output = io.StringIO()
        writer = csv.writer(output, quoting=csv.QUOTE_NONNUMERIC)
        for index, news in enumerate(self.news_list):
            if index == 0:
                writer.writerow(news.__dict__.keys())
            news_dict = deepcopy(news.__dict__)
            news_dict['published_at'] = news_dict['published_at']. \
                strftime(DATE_TIME_STR_FORMAT)
            writer.writerow(news_dict.values())
        return output.getvalue()
